@@ -109,7 +109,7 @@ class BotFramework(OutputChannel):
     async def send_text_message(
         self, recipient_id: Text, text: Text, **kwargs: Any
     ) -> None:
-        for message_part in text.split("\n\n"):
+        for message_part in text.strip().split("\n\n"):
             text_message = {"text": message_part}
             message = self.prepare_message(recipient_id, text_message)
             await self.send(message)
@@ -152,7 +152,6 @@ class BotFramework(OutputChannel):
     async def send_custom_json(
         self, recipient_id: Text, json_message: Dict[Text, Any], **kwargs: Any
     ) -> None:
-        # pytype: disable=attribute-error
         json_message.setdefault("type", "message")
         json_message.setdefault("recipient", {}).setdefault("id", recipient_id)
         json_message.setdefault("from", self.bot)
@@ -161,7 +160,6 @@ class BotFramework(OutputChannel):
         ).setdefault("alert", "true")
         json_message.setdefault("text", "")
         await self.send(json_message)
-        # pytype: enable=attribute-error
 
 
 class BotFrameworkInput(InputChannel):
@@ -176,9 +174,7 @@ class BotFrameworkInput(InputChannel):
         if not credentials:
             cls.raise_missing_credentials_exception()
 
-        # pytype: disable=attribute-error
         return cls(credentials.get("app_id"), credentials.get("app_password"))
-        # pytype: enable=attribute-error
 
     def __init__(self, app_id: Text, app_password: Text) -> None:
         """Create a Bot Framework input channel.
